#include "manager.h"
#include "gameobject.h"
#include "ship.h"
#include "graphics.h"
#include "constants.h"
#include "enemies.h"

#include <algorithm>
#include <functional>
#include <cstdlib>

using namespace std;

const double MinScrollSpeed = 0.015; // Units per tick (1/50 s)

void Manager::init()
{
	for (int i=0; i<8; i++) {
		mousebuttons[i] = false;
	}
	rm.Load();
	playerShip = 0;
	playerInfo.reset();
}

void Manager::render()
{
	glColor3f(0.6,0.6,0.6);
	glTranslatef(- scrollOffset, -4.5, 0);
	starfield.render();
	glEnable(GL_DEPTH_TEST);
	renderBackground();

	glColor3f(1,1,1);
	for (GameObjectIterator i = objects.begin(); i != objects.end(); i++) {
		(*i)->render();
	}
	// Show the mouse pos (debug)
	/*
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_LIGHTING);
	glEnable(GL_BLEND);
        glBlendFunc(GL_ONE, GL_SRC_ALPHA);
	DrawLuminousQuad(mousePosition, 0.2, 1,1,1);
	*/
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_BLEND);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glColor3f(1,1,1);
	DrawTextString(Vect3d(scrollOffset, 9,0));
	drawHitPoints();
}

static bool IsObjectDead(GameObject *obj) {
	return obj->isDead();
}

double Manager::getScrollSpeed() {
	return (MinScrollSpeed * scrollSpeed);
}

void Manager::runLogic()
{
	// Scroll.
	scrollOffset += getScrollSpeed();
	createScrolledObjects();
	calcMousePosition();
	// Run the actual objects' individual logic
	for (GameObjectIterator i = objects.begin(); i != objects.end(); i++) {
		(*i)->runLogic();
	}
	// Detect collisions
	collideDetect();
	
	// Take anything on the new list, add it to the current list.
	for (GameObjectIterator i = newobjects.begin(); i != newobjects.end(); i++) {
		objects.push_back(*i);
	}
	newobjects.clear();
	// Take anything on the dead list, and delete it.
	deleteList(deadobjects);
	// Take anything in the main list which is now dead, and add it to the dead list.
	for (GameObjectIterator i = objects.begin(); i != objects.end(); i++) {
		if ((*i)->isDead()) {
			deadobjects.push_back(*i);
		}
	}
	// Now remove them
	GameObjectIterator new_end =
                std::remove_if(objects.begin(), objects.end(), IsObjectDead);
        objects.erase(new_end, objects.end());
	if (playerShip != 0) {
		playerInfo.hitPoints = playerShip->hitPoints;
		if (playerShip->isDead()) {
			playerShip = 0;
			stateChangePending = true;
			playerDead = true;
		} else {
			playerLocation = playerShip->getLocation();
			if (scrollSpeed == 0) {
				if (((playerLocation.x - scrollOffset) > 0) && ! levelComplete) {
					// Level complete!
					stateChangePending = true;
					levelComplete = true;
					UpdateTextString("Level complete!");
				}
			}
		}
	}
	if (stateChangePending) {
		if (stateChangeTimer >0) {
			stateChangeTimer --;
		}
	}
	starfield.runLogic( *this);
	updatePlayerInfo();
}

// Create objects as they scroll in

void Manager::createScrolledObjects()
{
	int blockx =(int) scrollOffset + 7;
	if (blockx > maxblockx) {
		// Create the objects for this column
		for (int y=0; y<9; y++) {
			MapBlock & block = map.blockAt(blockx,y);
			GameObject * obj = 0;
			switch(block.item) {
				case MapBlock::TurretItem1:
				case MapBlock::TurretItem2:
				case MapBlock::TurretItem3:
					obj = new Turret(*this, block.item);
					break;
				case MapBlock::SaucerItem1:
				case MapBlock::SaucerItem2:
					obj = new Saucer(*this, block.item);
					break;
				default:
					break;
			}
			if (obj != 0) {
				obj->setLocation(Vect3d((double) blockx + 0.5, (double) y + 0.5, 0));
				addObject(obj);
			}
		}
		maxblockx = blockx;
		scrollSpeed = map.getSpeed(blockx);
	}
}

void Manager::addObject(GameObject * obj)
{
	// Add it to the newobjects vector
	newobjects.push_back(obj);
}
void Manager::deleteList(GameObjectList &list)
{
	for (GameObjectIterator i = list.begin(); i != list.end(); i++) {
		GameObject * obj = *i;
		delete obj;
	}
	list.clear();
}

void Manager::deleteAll()
{
	deleteList(newobjects);
	deleteList(deadobjects);
	deleteList(objects);
}

Manager::~Manager()
{
	deleteAll();
}

void Manager::initLevel()
{
	deleteAll();
	// init the map.
	map.init(playerInfo.level);
	starfield.init();
	// Create the ship.
	Ship * ship = new Ship(*this);
	addObject(ship);
	ship->setLocation(Vect3d(0,4.5,0));
	playerShip = ship;
	scrollOffset = 6.0;
	mouseScreenx = 1.0;
	mouseScreeny = 0.5;
	maxblockx = 0;
	scrollSpeed = 1;
	stateChangePending = false;
	stateChangeTimer = 150;
	levelComplete = false;
	UpdateTextString("Starting...");
	playerInfo.needsUpdate = true;
}

void Manager::updatePlayerInfo()
{
	if (playerInfo.needsUpdate) {
		char buf[100];
		sprintf(buf, "lev:%d lives:%d score:%06d", playerInfo.level, 
			playerInfo.lives, playerInfo.score);
		UpdateTextString(buf);
		playerInfo.needsUpdate = false;
	}
}

void Manager::setMousePosition(double x, double y)
{
	mouseScreenx = x;
	mouseScreeny = y;
}

void Manager::calcMousePosition()
{
	double x = mouseScreenx;
	double y = mouseScreeny;
	// x and y are in the range 0..1
	double viewportWidth = 12 / (CameraHeight / (CameraHeight - Ship::ShipHeight));
	double viewportHeight = viewportWidth * (3.0 / 4.0);
	x -= 0.5;
	y = 0.5 - y;
	x *= viewportWidth;
	y *= viewportHeight;
	// Make Y start at 0
	y += 4.5;
	mousePosition.x = x + scrollOffset;
	mousePosition.y = y;
	mousePosition.z = Ship::ShipHeight;
}

void Manager::renderBackground()
{
	glDisable(GL_LIGHTING);
	glDisable(GL_BLEND);
	glEnable(GL_TEXTURE_2D);
	rm.tiletexture.Bind();
	int minx = (int) scrollOffset - 6;
	int maxx = minx+12;
	for (int x=minx; x<=maxx; x++) {
		map.renderColumn(x, rm);
	}
}

void Manager::collideDetect()
{
	for (GameObjectIterator i = objects.begin(); i != objects.end(); i++) {
		GameObject * obj1 = *i;
		if (obj1->isCollidable()) 
			for (GameObjectIterator j = objects.begin(); j != objects.end(); j++) {
				GameObject * obj2 = *j;
				if (obj2->isProjectile()) {
					testCollision(obj1, obj2);
				}
			}
	}
}

void Manager::testCollision(GameObject * collidable, GameObject *projectile)
{
	double radiusSum = collidable->getRadius() + projectile->getRadius();
	Vect3d relPos = projectile->getLocation() - collidable->getLocation();
	relPos.z = 0;
	if (relPos.magnitudeSquared() < (radiusSum * radiusSum)) {
		collidable->handleCollision(projectile);
		projectile->handleCollision(collidable);
	}
}

void Manager::drawHitPoints()
{
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_BLEND);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glPushMatrix();
	// Top-left: 0,0
	glTranslatef(scrollOffset - 6.0, 9.0, 0);
	float healthEnd = (playerInfo.hitPoints / 100.0) * 3.0;
	glBegin(GL_QUADS);
	glColor3f(0.7,0,0);
	glVertex3f(0,0,0);
	glVertex3f(0,-0.5,0);
	glColor3f(0.5,1.0,0.5);
	glVertex3f(3,-0.5,0);
	glVertex3f(3,0,0);
	// Black quad over the empty bit.
	glColor3f(0,0,0);
	glVertex3f(healthEnd,0,0);
	glVertex3f(healthEnd,-0.5,0);
	glVertex3f(3,-0.5,0);
	glVertex3f(3,0,0);
	glEnd();
	
	glColor3f(1,1,1);
	glBegin(GL_LINE_LOOP);
	glVertex3f(0,0,0);
	glVertex3f(3,0,0);
	glVertex3f(3,-0.5,0);
	glVertex3f(0,-0.5,0);
	glEnd();
	glPopMatrix();
}


const int NumStars = 40;
/* Starfield stuff */

StarField::StarField() 
{ 
	// Create stars.
	init();	
}
void StarField::runLogic(Manager &m)
{
	// Move any stars which are off the LHS of the screen, 
	// on to the right.
	for (StarList::iterator i = stars.begin(); i != stars.end(); i++) {
		Star &s = *i;
		double xrange = (12 - s.location.z);
		double xtrans = s.location.x - m.getScrollOffset();
		if (xtrans < (-0.5 * xrange)) {
			s.location.x += xrange;
		}
	}
	
}

void StarField::render()
{
	glDisable(GL_LIGHTING);
	glDisable(GL_BLEND);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_DEPTH_TEST);
        glPointSize(2.0);
	glBegin(GL_POINTS);
	for (StarList::iterator i = stars.begin(); i != stars.end(); i++) {
		Star &s = *i;
		glColor3f(s.brightness, s.brightness, s.brightness);
		glVertex3f(s.location.x, s.location.y, s.location.z);
	}
	glEnd();

}

// (re) init the stars.
void StarField::init()
{
	stars.clear();
	stars.resize(NumStars);
	for (StarList::iterator i = stars.begin(); i != stars.end(); i++) {
		Star &s = *i;
		double z = -6 - (rand() % 24);
		double xrange = (12 - z);
		double yrange = xrange * (3.0/4.0);
		double x = rand();
		double y = rand();
		x = x / RAND_MAX;
		y = y / RAND_MAX;
		x -= 0.5;
		y -= 0.5;
		x *= xrange;
		y *= yrange;
		y += 4.5; // compensate for translation.
		s.location = Vect3d(x,y,z);
		s.brightness = 0.3 + (((double) rand()) / RAND_MAX) * 0.7;
	}
}


